#include "psolver.h"
#include <string.h>
#include <math.h>
#include <limits.h>

psolver::psolver (int np,int mr)
{
  //  number of processors
  nproc=np;
  //  my rank = processor id
  myrank=mr;
  //  number of executed domain
  ndom=0;
  
  //  type of domain decomposition method
  tdd = (domdectype) 0;
  //  type of solver of reduced system of equations
  trssol = (redsystsolver) 0;
  //  type of storage of the matrix of the reduced system
  rsmstor = (storagetype) 0;
  //  type of mesh description
  md = (meshdescription) 0;


  zero=1.0e-15;
  
  //  local to global array
  ltg = NULL;
  //  correspondence between subdomains and processors
  domproc = NULL;
  
  //  list of numbers of nodes on subdomain
  nndom = NULL;
  //  list of numbers of elements on subdomain
  nedom = NULL;
  //  list of numbers of stored matrix entries on subdomain
  negmdom = NULL;

  
  //  object containing the Schur complement method
  schcom = NULL;
  //  object containing the one-level FETI method
  f1 = NULL;
  //  object containing the FETI-DP method
  dpf = NULL;
  //  object containing the layered plate problem
  lpp = NULL;
  //  object containing the parallel conjugate gradient method
  parcg = NULL;
  //  object containing the boundary version of the parallel conjugate gradient method
  boundparcg = NULL;
  
  //  parallel topology
  ptop=NULL;

  ssle = new slesolv();
}

psolver::~psolver()
{

}

/**
   function initiates parallel solver
   
   @param top - pointer to the general topology
   
   JK
*/
void psolver::initiate (gtopology *top)
{
  switch (tdd){
  case schurcompldd:{
    schcom = new schurcompl (nproc,myrank,ndom);
        
    schcom->trssol    = trssol;
    schcom->rsmstor   = rsmstor;
    schcom->ssle->initiate (ssle);
    schcom->dmstor    = dmstor;
    schcom->nicgsch   = nicg;
    schcom->errcgsch  = errcg;
    schcom->zero      = zero;

    break;
  }
  case fetidd:{
    f1 = new feti1 (nproc,myrank,ndom);
    
    f1->enrbm = enrbm;
    f1->lithr = lithr;
    f1->nicg  = nicg;
    f1->errcg = errcg;
    f1->zero  = zero;
    f1->fetiprecond = fetiprecond;
    break;
  }
  case dpfetidd:{
    dpf = new dpfeti (nproc,myrank,ndom);

    dpf->rsmstor = rsmstor;
    dpf->ssle->initiate (ssle);
    dpf->nicgdpfeti  = nicg;
    dpf->errcgdpfeti = errcg;
    dpf->zero  = zero;

    break;
  }
  case parconjuggrad:{
    parcg = new parcongrad (nproc,myrank,ndom);
    
    parcg->ni = nicg;
    parcg->err = errcg;
    break;
  }
  case boundparconjuggrad:{
    boundparcg = new boundparcongrad (nproc,myrank,ndom);
    boundparcg->ni = nicg;
    boundparcg->err = errcg;
    break;
  }
  case layered_plate:{
    lpp = new lplate (nproc,myrank,ndom,top);
    
    lpp->nicg = nicg;
    lpp->errcg = errcg;
    break;
  }
  default:{
    fprintf (stderr,"\n\n (%d) unknown domain decomposition solver is required",myrank);
    fprintf (stderr,"\n in function initiate (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  //  allocation of parallel topology
  ptop = new partop (nproc,myrank,ndom,md);
}

/**
   function reads basic facts about parallel solver of algebraic equations
   
   @param in - input stream
   @param nd - number of subdomain
   @param tsm - storage typo of subdomain matrix
   @param mespr - message printing (yes if 1)

   9.10.2003, JK
*/
void psolver::read (XFILE *in,long nd,storagetype tsm,long mes)
{
  //  message printing
  mespr=mes;
  
  //  number of subdomain
  ndom=nd;
  
  //  type of storage of matrix of subdomains
  dmstor = tsm;
  
  //  type of mesh description
  xfscanf (in,"%k%m","md",&meshdescription_kwdset,(int*)&md);

  //  type of domain decomposition solver
  xfscanf (in,"%d",(int*)&tdd);
  
  
  switch (tdd){
    // ************************************************************************************
    //  SCHUR COMPLEMENT METHOD - PRIMAL DOMAIN DECOMPOSITION METHOD - CONDENSATION METHOD
    // ************************************************************************************
  case schurcompldd:{
    if (mespr==1)  fprintf (stdout,"\n primal domain decomposition (Schur complement method) is used");
    
    //  type of reduced solver
    xfscanf (in,"%d",(int*)&trssol);
    switch (trssol){
    case master_sol:{
      if (mespr==1)  fprintf (stdout,"\n reduced system of equations is solved sequentially on master processor");
      xfscanf (in,"%d",(int*)&rsmstor);
      
      ssle->read (in,mespr);
      
      break;	
    }
    case paral_sol:{
      if (mespr==1)  fprintf (stdout,"\n reduced system of equations is solved in parallel");
      xfscanf (in,"%ld %le",&nicg,&errcg);
      break;
    }
    default:{
      fprintf (stderr,"\n\n PROC %d: unknown reduced system solver is required",myrank);
      fprintf (stderr,"\n in function pprobdesc::read (file %s, line %d).\n",__FILE__,__LINE__);
    }
    }
    break;
  }
    // **************************************************************************************
    //  FINITE ELEMENT TEARING AND INTERCONNECTING METHOD - DUAL DOMAIN DECOMPOSITION METHOD
    // **************************************************************************************
  case fetidd:{
    if (mespr==1)  fprintf (stdout,"\n dual domain decomposition (Finite Element Tearing and Interconnecting method) is used");
    xfscanf (in,"%ld %lf",&enrbm,&lithr);
    fprintf (stdout,"\n proc %d  lithr %e",myrank,lithr);
    xfscanf (in,"%ld %le",&nicg,&errcg);
    xfscanf (in,"%ld",&fetiprecond);
    switch (fetiprecond){
    case nofetiprecond:{
      fprintf (stdout,"\n No preconditioner is used for FETI method");
      break;}
    case lumped:{
      fprintf (stdout,"\n Lumped preconditioner is used for FETI method");
    }
    case dirichlet:{
      fprintf (stdout,"\n Lumped preconditioner is used for FETI method");
      break;
    }
    }
    break;
  }
    // ***************************************************************
    //  DUAL-PRIMAL FINITE ELEMENT TEARING AND INTERCONNECTING METHOD
    // ***************************************************************
  case dpfetidd:{
    if (mespr==1)  fprintf (stdout,"\n dual-primal finite element tearing and interconnecting method is used");
    
    //  type of reduced solver
    xfscanf (in,"%d",(int*)&trssol);
    switch (trssol){
    case master_sol:{
      if (mespr==1)  fprintf (stdout,"\n reduced system of equations is solved sequentially on master processor");
      xfscanf (in,"%d",(int*)&rsmstor);
      ssle->read (in,mespr);
      xfscanf (in,"%ld %le",&nicg,&errcg);
      break;	
    }
    default:{
      fprintf (stderr,"\n\n PROC %d: unknown reduced system solver is required",myrank);
      fprintf (stderr,"\n in function pprobdesc::read (file %s, line %d).\n",__FILE__,__LINE__);
    }
    }
    
    break;
  }
    // *************************************
    //  PARALLEL CONJUGATE GRADIENT METHOD
    // *************************************
  case parconjuggrad:{
    if (mespr==1)  fprintf (stdout,"\n parallel conjugate gradient method is used");
    
    xfscanf (in,"%ld %le",&nicg,&errcg);
    
    break;
  }
    // *******************************************************
    // BOUNDARY VERSION OF  PARALLEL CONJUGATE GRADIENT METHOD
    // *******************************************************
  case boundparconjuggrad:{
    if (mespr==1)  fprintf (stdout,"\n boundary version of parallel conjugate gradient method is used");
    
    xfscanf (in,"%ld %le",&nicg,&errcg);
    break;
  }
   


  case layered_plate:{
    
    if (mespr==1)  fprintf (stdout,"\n layered linear plate problem is solved with help of orthonormalization of constraints");
    
    //  data about solver of system of linear equations
    ssle->read (in,mespr);
    
    //  data about conjugate gradient method
    xfscanf (in,"%ld %le",&nicg,&errcg);
    
    break;
  }
    
    // ****************************************************************
    // ****************************************************************
  default:{
    fprintf (stderr,"\n PROC %d: unknown domain decomposition solver is required",myrank);
    fprintf (stderr,"\n in function psolv::read (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
}

/**
   function prints facts about parallel solver
   
   9.10.2003, JK
*/
void psolver::print (FILE *out)
{
  //  type of mesh description
  fprintf (out,"%d\n",md);
  //  type of domain decomposition solver
  fprintf (out,"%d\n", tdd);
  
  switch (tdd){
    // ************************************************************************************
    //  SCHUR COMPLEMENT METHOD - PRIMAL DOMAIN DECOMPOSITION METHOD - CONDENSATION METHOD
    // ************************************************************************************
  case schurcompldd:{
    
    //  type of reduced solver
    fprintf (out,"%d ",(int)trssol);
    
    switch (trssol){
    case master_sol:{
      fprintf (out,"%d\n",(int)rsmstor);
      
      ssle->print (out);
      fprintf (out,"\n");
      
      break;	
    }
    case paral_sol:{
      fprintf (out,"%ld %e\n",nicg,errcg);
      break;
    }
    default:{
      fprintf (stderr,"\n\n PROC %d: unknown reduced system solver is required",myrank);
      fprintf (stderr,"\n in function pprobdesc::read (file %s, line %d).\n",__FILE__,__LINE__);
    }
    }
    
    break;
  }
    // **************************************************************************************
    //  FINITE ELEMENT TEARING AND INTERCONNECTING METHOD - DUAL DOMAIN DECOMPOSITION METHOD
    // **************************************************************************************
  case fetidd:{
    fprintf (out,"%ld %e ",enrbm,lithr);
    fprintf (out,"%ld %e\n",nicg,errcg);
    fprintf (out,"%ld\n",fetiprecond);
    switch (fetiprecond){
    case nofetiprecond:{
      fprintf (stdout,"\n No preconditioner is used for FETI method");
      break;}
    case lumped:{
      fprintf (stdout,"\n Lumped preconditioner is used for FETI method");
    }
    case dirichlet:{
      fprintf (stdout,"\n Lumped preconditioner is used for FETI method");
      break;
    }
    }
    break;
  }
    // ***************************************************************
    //  DUAL-PRIMAL FINITE ELEMENT TEARING AND INTERCONNECTING METHOD
    // ***************************************************************
  case dpfetidd:{
    //  type of reduced solver
    fprintf (out,"%d\n",(int)trssol);
    switch (trssol){
    case master_sol:{
      fprintf (out,"%d\n",(int)rsmstor);
      ssle->print (out);
      fprintf (out,"%ld %e\n",nicg,errcg);
      break;	
    }
    default:{
      fprintf (stderr,"\n\n PROC %d: unknown reduced system solver is required",myrank);
      fprintf (stderr,"\n in function pprobdesc::print (file %s, line %d).\n",__FILE__,__LINE__);
    }
    }
    
    break;
  }
    // *************************************
    //  PARALLEL CONJUGATE GRADIENT METHOD
    // *************************************
  case parconjuggrad:{
    
    fprintf (out,"%ld %le",nicg,errcg);
    
    break;
  }
    // *******************************************************
    // BOUNDARY VERSION OF  PARALLEL CONJUGATE GRADIENT METHOD
    // *******************************************************
  case boundparconjuggrad:{
    
    fprintf (out,"%ld %le",nicg,errcg);
    break;
  }
    
  case layered_plate:{
    //  data about solution strategy
    ssle->print (out);
    fprintf (out,"\n");
    
    //  data about conjugate gradient method
    fprintf (out,"%ld %e\n",nicg,errcg);
    
    break;
  }
    
  default:{
    fprintf (stderr,"\n PROC %d: unknown domain decomposition solver is required",myrank);
    fprintf (stderr,"\n in function pprobdesc::read (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  fprintf (out,"\n");
  
}


/**
   function reads local to global correspondence of nodes
   
   @param in - input file
   @param top - pointer to the sequential general toplogy
   @param out - output file (used for auxiliary uotputs)
   
   JK, 20.1.2002
*/
void psolver::read_ltg (XFILE *in,gtopology *top,FILE *out)
{
  long i;
  
  ltg = new long [top->nn];
  
  for (i=0;i<top->nn;i++){
    xfscanf (in,"%ld",ltg+i);
    ltg[i]--;
    
    /*
    switch (tdd){
    case schurcompldd:{
      xfscanf (in,"%ld",ltg+i);
      if (md == all_nodes)
	ltg[i]=abs(ltg[i]);
      top->gnodes[i].ai = ltg[i];
      break;
    }
    case fetidd:{
      xfscanf (in,"%ld",ltg+i);
      if (md == all_nodes)
	ltg[i]=abs(ltg[i]);
      top->gnodes[i].ai = ltg[i];
      break;
    }
    case dpfetidd:{
      xfscanf (in,"%ld",ltg+i);
      if (md == all_nodes)
	ltg[i]=abs(ltg[i]);
      if (ltg[i]>=0)   top->gnodes[i].ai = 0;
      if (ltg[i]<0)    top->gnodes[i].ai = 0-ltg[i];
      break;
    }
    case parconjuggrad:{
      xfscanf (in,"%ld",ltg+i);
      if (md == all_nodes)
	ltg[i]=abs(ltg[i]);
      top->gnodes[i].ai = ltg[i];
      break;
    }
    case boundparconjuggrad:{
      xfscanf (in,"%ld",ltg+i);
      if (md == all_nodes)
	ltg[i]=abs(ltg[i]);
      top->gnodes[i].ai = ltg[i];
      break;
    }
    case layered_plate:{
      break;
    }
    default:{
      par_print_err(myrank,"unknown type of domain decomposition is required",__FILE__,__LINE__,__func__);
      abort ();
    }
    }
    
    ltg[i]--;
    */
  }
  
}

/**
   function establishes correspondence among processors and subdomains
   
   let i denotes number of processor and j denotes number of subdomain
   then domproc[i]=j
   
   JK
*/
void psolver::procdomcorr ()
{
  long i,j;
  MPI_Status stat;
  
  if (myrank==0){
    domproc = new long [nproc];
    domproc[0]=ndom;
    for (i=1;i<nproc;i++){
      MPI_Recv (&j,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);

      //fprintf (stderr,"\n i=%ld   j=%ld     stat %ld",i,j,stat.MPI_TAG);
      
      domproc[stat.MPI_TAG]=j;
    }
  }
  else{
    MPI_Send (&ndom,1,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);

  /*
  if (myrank==0){
    domproc = new long [nproc+1];
    domproc[ndom]=0;
    for (i=1;i<nproc;i++){
      MPI_Recv (&j,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      domproc[j]=stat.MPI_TAG;
    }
    for (i=0;i<nproc;i++){
      if (domproc[i]==0){
	domproc[nproc]=i;  break;
      }
    }
  }
  else{
    MPI_Send (&ndom,1,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  */
}


void psolver::nodesplit (gtopology *top,FILE *out)
{
  /*  stare
  ptop->nodesplit_bnodes (top,domproc,out);
  ptop->corner_nodes (top,out);
  */

  ptop->initiation (top,ltg);
  ptop->numbers_of_all_nodes_on_subdomains (domproc,out);
  ptop->compute_multiplicity (ltg,domproc,out);
  ptop->find_boundary_nodes (ltg,domproc,out);
  ptop->boundary_nodes_on_master (top,domproc,out);
}


/**
   function generates ordering of unknowns with respect
   to the selected type of domain decomposition method
   
   @param top - pointer to the sequential general topology
   @param out - output file (used for auxiliary output)
   
   JK, 11.7.2008 - revised
*/
long psolver::ordering (gtopology *top,FILE *out)
{
  long i,ndof;
  long *schurnodes,*fetinodes;
  
  /*
  selnod = new selnodes (nproc,myrank,ndom,top->nn,ltg,md,out);

  selnod->number_of_selected_nodes (domproc,out);
  selnod->nodes_on_master (domproc,out);
  selnod->node_multiplicity (out);
  selnod->number_all_dofs (top,domproc,out);
  selnod->ndofn_on_master (top,domproc,out);
  selnod->dof_indicators (top,domproc,out);
  selnod->schur_ordering (out);

  selnod->group_local_nodes (out);
  selnod->dof_feti (out);
  selnod->number_contrib (domproc,out);
  selnod->contrib_dofs (top,domproc,out);
  */





  
  /*
  ptop->initiation (top,ltg);
  ptop->numbers_of_all_nodes_on_subdomains (top,domproc,out);
  ptop->compute_multiplicity (top,domproc,out);
  ptop->find_boundary_nodes (top,domproc,out);
  ptop->boundary_nodes_on_master (top,domproc,out);
  ptop->number_of_bdofs_on_nodes (top,domproc,out);
  ptop->code_numbers_on_master (top,domproc,out);
   


  //ptop->corner_detection(top,domproc,out);
  
  // assemble gnn

  ptop -> numbers_of_all_dofs_on_subdomains (top,domproc,out);
  ptop -> assemble_gnn(top,domproc,out);
  ptop -> assemble_pgcn(top,domproc,out);
  ptop -> assemble_gcnd(top,domproc,out);
  */

  switch (tdd){
    // **************************
    //  SCHUR COMPLEMENT METHOD
    // **************************
  case schurcompldd:{
    /*
    ptop->initiation (top,ltg);
    ptop->numbers_of_all_nodes_on_subdomains (top,domproc,out);
    switch (md){
    case all_nodes:{
      ptop->compute_multiplicity (top,domproc,out);
      ptop->find_boundary_nodes (top,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      break;
    }
    case bound_nodes:{
      ptop->find_boundary_nodes (top,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      ptop->coarse_local_nodes ();
      ptop->compute_multiplicity (top,domproc,out);
      break;
    }
    default:{
      fprintf (stderr,"\n\n unknown type of mesh type is required in function");
      fprintf (stderr,"\n ordering (file %s, line %d)\n",__FILE__,__LINE__);
    }
    }


    ptop->number_of_bdofs_on_nodes (top,domproc,out);
    ptop->code_numbers_on_master (top,domproc,out);
    
    
    schcom->subdomain_ordering (top,ptop);
    schcom->coarse_problem_ordering (ptop,out);
    
    ndof=schcom->ndof;
    */
    //ptop->dof_multiplicity (top,out);


    ptop->initiation (top,ltg);
    ptop->numbers_of_all_nodes_on_subdomains (domproc,out);
    ptop->compute_multiplicity (ltg,domproc,out);
    ptop->find_boundary_nodes (ltg,domproc,out);
    ptop->rewrite_ltg (ltg);


    schcom->subdomain_ordering (top,ltg);
    ndof=schcom->ndof;
    
    schurnodes = new long [top->nn];
    for (i=0;i<top->nn;i++){
      if (ltg[i]>-1)
	schurnodes[i]=ltg[i];
      else
	schurnodes[i]=-1;
    }

    selnodschur = new selnodes (nproc,myrank,ndom,top->nn,schurnodes,md,out,mespr);
    selnodschur->number_of_selected_nodes (domproc,out);
    selnodschur->nodes_on_master (domproc,out);
    selnodschur->node_multiplicity (out);

    selnodschur->number_all_dofs (top,domproc,out);
    selnodschur->ndofn_on_master (top,domproc,out);
    selnodschur->dof_indicators (top,domproc,out);
    selnodschur->schur_ordering (top,out);
    
    schcom->initiate (selnodschur);
    
    
    //  this must be used for sparse direct solver
    for (i=0;i<top->nn;i++){
      top->gnodes[i].ai=ltg[i];
    }
    
    break;
  }
    //  FETI METHOD
  case fetidd:{
    /*
    ptop->initiation (top,ltg);
    ptop->numbers_of_all_nodes_on_subdomains (top,domproc,out);
    switch (md){
    case all_nodes:{
      ptop->compute_multiplicity (top,domproc,out);
      ptop->find_boundary_nodes (top,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      ptop->coarse_local_nodes ();
      break;
    }
    case bound_nodes:{
      ptop->find_boundary_nodes (top,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      ptop->coarse_local_nodes ();
      ptop->compute_multiplicity (top,domproc,out);
      break;
    }
    default:{
      fprintf (stderr,"\n\n unknown type of mesh type is required in function");
      fprintf (stderr,"\n ordering (file %s, line %d)\n",__FILE__,__LINE__);
    }
    }
    ptop->number_of_bdofs_on_nodes (top,domproc,out);
    ptop->code_numbers_on_master (top,domproc,out);
    
    //ptop->coarse_local_nodes ();
    ptop->sort_nodes (out);
    
    f1->subdomain_ordering (top,out);
    f1->coarse_problem_ordering (top,ptop,domproc,out);
    */


    ptop->initiation (top,ltg);
    ptop->numbers_of_all_nodes_on_subdomains (domproc,out);
    ptop->compute_multiplicity (ltg,domproc,out);
    ptop->find_boundary_nodes (ltg,domproc,out);
    ptop->rewrite_ltg (ltg);

    f1->subdomain_ordering (top,ltg,out);
    fetinodes = new long [top->nn];

    fprintf (out,"\n\n kontrola fetinodes");
    for (i=0;i<top->nn;i++){
      if (ltg[i]>-1)
	fetinodes[i]=ltg[i];
      else
	fetinodes[i]=-1;
      
      fprintf (out,"\n fetinodes %ld   %ld    ltg %ld",i,fetinodes[i],ltg[i]);
    }
    

    selnodfeti = new selnodes (nproc,myrank,ndom,top->nn,fetinodes,md,out,mespr);
    selnodfeti->number_of_selected_nodes (domproc,out);
    selnodfeti->nodes_on_master (domproc,out);
    selnodfeti->node_multiplicity (out);
    selnodfeti->number_all_dofs (top,domproc,out);
    selnodfeti->ndofn_on_master (top,domproc,out);
    selnodfeti->dof_indicators (top,domproc,out);
    selnodfeti->group_local_nodes (out);
    selnodfeti->dof_feti (out);
    selnodfeti->number_contrib (domproc,out);
    selnodfeti->contrib_dofs (top,domproc,out);
    
    // dofmultip for scaling
    selnodfeti->dof_multiplicity (out);
      
    f1->initiate (selnodfeti,out);

    ndof=f1->ndof;
    break;
  }
    //
    //  FETI-DP METHOD
    //
  case dpfetidd:{
    
    //dpf->cornodedetection (top,ltg,domproc,out);
    // 17.04.07 - toto je platne
    // hledani vertex, edge a internal nodes
    //ptop->identification_node(top,domproc,out);
    
    /*
    ptop->initiation (top,ltg);
    ptop->numbers_of_all_nodes_on_subdomains (top,domproc,out);
    switch (md){
    case all_nodes:{
      ptop->compute_multiplicity (top,domproc,out);
      ptop->find_boundary_nodes (top,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      break;
    }
    case bound_nodes:{
      ptop->find_boundary_nodes (top,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      ptop->coarse_local_nodes ();
      ptop->compute_multiplicity (top,domproc,out);
      break;
    }
    default:{
      fprintf (stderr,"\n\n unknown type of mesh type is required in function");
      fprintf (stderr,"\n ordering (file %s, line %d)\n",__FILE__,__LINE__);
    }
    }


    ptop->corner_detection (top,domproc,ltg,out);
    */


    /*
    dpf->dpfetiordering (top,ltg);
    dpf->globcnnum_dpfeti (top,ltg,domproc,out);
    ndof=dpf->ndof;
    */
    //ptop->dof_multiplicity (top,out);
    
    
    
    
    
    
    /*
    ptop->initiation (top,ltg);
    ptop->numbers_of_all_nodes_on_subdomains (top,domproc,out);
    switch (md){
    case all_nodes:{
      ptop->compute_multiplicity (top,domproc,out);
      ptop->find_boundary_nodes (top,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      break;
    }
    case bound_nodes:{
      ptop->find_boundary_nodes (top,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      ptop->coarse_local_nodes ();
      ptop->compute_multiplicity (top,domproc,out);
      break;
    }
    default:{
      fprintf (stderr,"\n\n unknown type of mesh type is required in function");
      fprintf (stderr,"\n ordering (file %s, line %d)\n",__FILE__,__LINE__);
    }
    }


    ptop->corner_detection (top,domproc,ltg,out);

    
    fprintf (out,"\n\n kontrola ltg \n");
    for (i=0;i<top->nn;i++){
      fprintf (out,"\n %ld",ltg[i]+1);
    }
    fprintf (out,"\n konec kontroly");
    */

    fprintf (out,"\n\n kontrola pole ltg");
    for (i=0;i<top->nn;i++){
      fprintf (out,"\n %ld ltg %ld",i,ltg[i]);
    }



    ptop->initiation (top,ltg);
    ptop->numbers_of_all_nodes_on_subdomains (domproc,out);
    ptop->compute_multiplicity (ltg,domproc,out);
    ptop->find_boundary_nodes (ltg,domproc,out);
    //ptop->boundary_nodes_on_master (top,domproc,out);
    //ptop->corner_detection (top,domproc,ltg,out); 
    //fprintf (out,"\n\n kontrola md = %d",md);
    // rewrite mesh descripton
    //md = bound_nodes;
    //fprintf (out,"\n\n kontrola md = %d",md);
    
    long *check;
    check = new long[1];
    check[0] = -1;
    ptop->check_corner_ltg(top,domproc,ltg,check,out);
    if (check[0] == 1){
      fprintf (out,"\n\n Array ltg was changed into format bound_nodes with corner nodes\n");
      // rewrite mesh descripton
      md = bound_nodes;
    }
    
    //  this must be used for sparse direct solver


    for (i=0;i<top->nn;i++){
      if (ltg[i]>-2)
	top->gnodes[i].ai=-1;
      else
	top->gnodes[i].ai = 0-ltg[i]-2;
    }

    
    
    fprintf (out,"\n\n kontrola pole ltg");
    for (i=0;i<top->nn;i++){
      fprintf (out,"\n ltg %ld   %ld",i+1,ltg[i]);
    }
    fprintf (out,"\n");
    
    
    dpf->dpfetiordering (top,ltg);
    ndof=dpf->ndof;
    
    fetinodes = new long [top->nn];

    fprintf (out,"\n\n kontrola fetinodes");
    for (i=0;i<top->nn;i++){
      if (ltg[i]>-1)
	fetinodes[i]=ltg[i];
      else
	fetinodes[i]=-1;
      
      fprintf (out,"\n fetinodes %ld   %ld    ltg %ld",i,fetinodes[i],ltg[i]);
    }
    

    selnodfeti = new selnodes (nproc,myrank,ndom,top->nn,fetinodes,md,out,mespr);

    selnodfeti->number_of_selected_nodes (domproc,out);
    selnodfeti->nodes_on_master (domproc,out);
    selnodfeti->node_multiplicity (out);

    selnodfeti->number_all_dofs (top,domproc,out);
    selnodfeti->ndofn_on_master (top,domproc,out);
    selnodfeti->dof_indicators (top,domproc,out);
    selnodfeti->group_local_nodes (out);

    selnodfeti->dof_feti (out);

    selnodfeti->number_contrib (domproc,out);
    selnodfeti->contrib_dofs (top,domproc,out);
    
    
    fprintf (out,"\n\n\n KONEC FETI, ZACATEK SCHURA \n\n\n");
    
    schurnodes = new long [top->nn];
    for (i=0;i<top->nn;i++){
      if (ltg[i]<-1)
	schurnodes[i]=0-ltg[i]-2;
      else
	schurnodes[i]=-1;
    }

    fprintf (out,"\n\n kontrola schurnodes");
    for (i=0;i<top->nn;i++){
      /*
	if (ltg[i]>-1)
	schurnodes[i]=ltg[i];
      else
	schurnodes[i]=-1;
      */
      fprintf (out,"\n schurnodes %ld   %ld    ltg %ld",i,schurnodes[i],ltg[i]);
    }
    

    selnodschur = new selnodes (nproc,myrank,ndom,top->nn,schurnodes,md,out,mespr);
    selnodschur->number_of_selected_nodes (domproc,out);
    selnodschur->nodes_on_master (domproc,out);
    selnodschur->node_multiplicity (out);
    selnodschur->number_all_dofs (top,domproc,out);

    selnodschur->ndofn_on_master (top,domproc,out);
    selnodschur->dof_indicators (top,domproc,out);
    selnodschur->schur_ordering (top,out);

    
    dpf->initiate (selnodschur,selnodfeti);


    break;
  }
  case parconjuggrad:{
    ptop->initiation (top,ltg);
    ptop->numbers_of_all_nodes_on_subdomains (domproc,out);
    switch (md){
    case all_nodes:{
      ptop->compute_multiplicity (ltg,domproc,out);
      ptop->find_boundary_nodes (ltg,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      break;
    }
    case bound_nodes:{
      ptop->find_boundary_nodes (ltg,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      ptop->coarse_local_nodes ();
      ptop->compute_multiplicity (ltg,domproc,out);
      break;
    }
    default:{
      fprintf (stderr,"\n\n unknown type of mesh type is required in function");
      fprintf (stderr,"\n ordering (file %s, line %d)\n",__FILE__,__LINE__);
    }
    }
    ptop->numbers_of_all_dofs_on_subdomains (top,domproc,out);
    
    ptop->assemble_gnn(top,domproc,out);
    ptop->assemble_pgcn(top,domproc,out);
    ptop->assemble_gcnd(top,domproc,out);
    
    
    parcg->subdomain_ordering (ptop,top,out);
    ndof=parcg->ndof;
    ptop->dof_multiplicity (top,out);
    break;
  }
  case boundparconjuggrad:{
    /*
    ptop->initiation (top,ltg);
    ptop->numbers_of_all_nodes_on_subdomains (domproc,out);
    switch (md){
    case all_nodes:{
      ptop->compute_multiplicity (ltg,domproc,out);
      ptop->find_boundary_nodes (ltg,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      break;
    }
    case bound_nodes:{
      ptop->find_boundary_nodes (ltg,domproc,out);
      ptop->boundary_nodes_on_master (top,domproc,out);
      ptop->coarse_local_nodes ();
      ptop->compute_multiplicity (ltg,domproc,out);
      break;
    }
    default:{
      fprintf (stderr,"\n\n unknown type of mesh type is required in function");
      fprintf (stderr,"\n ordering (file %s, line %d)\n",__FILE__,__LINE__);
    }
    }
    
    ptop->number_of_bdofs_on_nodes (top,domproc,out);
    ptop->code_numbers_on_master (top,domproc,out);
    
    ptop->numbers_of_all_dofs_on_subdomains (top,domproc,out);
    
   
    boundparcg ->coarse_problem_ordering (domproc,top,ptop,out);
    
    ndof=boundparcg->ndof;
    ptop->dof_multiplicity (top,out);
    */


    ptop->initiation (top,ltg);
    ptop->numbers_of_all_nodes_on_subdomains (domproc,out);
    ptop->compute_multiplicity (ltg,domproc,out);
    ptop->find_boundary_nodes (ltg,domproc,out);
    ptop->rewrite_ltg (ltg);

    boundparcg->subdomain_ordering (ptop,top,out);
    
    
    schurnodes = new long [top->nn];
    
    for (i=0;i<top->nn;i++){
      if (ltg[i]>-1)
	schurnodes[i]=ltg[i];
      else
	schurnodes[i]=-1;
    }
    
    selnodschur = new selnodes (nproc,myrank,ndom,top->nn,schurnodes,md,out,mespr);
    selnodschur->number_of_selected_nodes (domproc,out);
    selnodschur->nodes_on_master (domproc,out);
    selnodschur->node_multiplicity (out);
    selnodschur->number_all_dofs (top,domproc,out);
    
    selnodschur->ndofn_on_master (top,domproc,out);
    selnodschur->dof_indicators (top,domproc,out);
    selnodschur->schur_ordering (top,out);
    
    boundparcg->initiate (selnodschur,ptop,top,out);
    ndof=boundparcg->ndof;
    
    
    
    
    break;
  }
  case layered_plate:{
    lpp->ndof = top->gencodnum ();
    lpp->globcnnum_lpp (top,domproc,out);
    ndof=lpp->ndof;
    break;
  }
  default:{
    par_print_err(myrank,"unknown type of domain decomposition solver is required",__FILE__,__LINE__,__func__);
  }
  }
  
  /*
  long i;
  for (i=0;i<top->nn;i++){
    
    switch (tdd){
    case schurcompldd:{
      top->gnodes[i].ai = ltg[i]+1;
      break;
    }
    case fetidd:{
      top->gnodes[i].ai = ltg[i]+1;
      break;
    }
    case dpfetidd:{
      if (ltg[i]+1>=0)   top->gnodes[i].ai = 0;
      if (ltg[i]+1<0)    top->gnodes[i].ai = 0-ltg[i]-1;
      break;
    }
    case parconjuggrad:{
      top->gnodes[i].ai = ltg[i]+1;
      break;
      }
    case layered_plate:{
      break;
    }
    default:{
      fprintf (stderr,"\n\n unknown type of domain decomposition is required in function read (file %s, line %d).\n",__FILE__,__LINE__);
      abort ();
    }
    }
    
  }
  */
  
  
  //ptop->dof_multiplicity (top,out);

  return ndof;
}

/**
   function assembles constraint matrix
   function is used in layered static analysis
   
   @param th - array containing thicknesses in nodes
   @param out - output stream
   
   JK, 4.3.2003
*/
void psolver::constr_mat (double *th,FILE *out)
{
  switch (tdd){
  case layered_plate:{
    lpp->constraintmat (th,domproc,out);
    break;
  }
  default:{
    fprintf (stderr,"\n\n (%d) unknown domain decomposition solver is required",myrank);
    fprintf (stderr,"\n in function constr_mat (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }

}

/**
   function solves system of linear algebraic equations in parallel
   
   @param top - pointer to the general topology describing appropriate subdomain
   @param gm - pointer to the %matrix of subdomain
   @param lhs - array containing solution
   @param rhs - array containing right hand side
   @param out - output file (for auxiliary output)
   @param mespr - message printing indicator
   
   JK, revised 24.7.2008
*/
void psolver::par_linear_solver (gtopology *top,gmatrix *gm,
				 double *lhs,double *rhs,FILE *out,long mespr)
{
  switch (tdd){
  case schurcompldd:{
    schcom->solve_system (top,gm,domproc,lhs,rhs,out);
    break;
  }
  case fetidd:{
    f1->solve_system (top,gm,domproc,lhs,rhs,out);
    break;
  }
  case dpfetidd:{
    dpf->solve_system (top,gm,domproc,lhs,rhs,out,mespr);
    break;
  }
  case parconjuggrad:{
    parcg->solve_system (ptop,top,gm,domproc,lhs,rhs,out,0);
    break;
  }
  case boundparconjuggrad:{
    boundparcg->solve_system (ptop,top,gm,domproc,lhs,rhs,out,0);
    break;
  }
  case layered_plate:{
    lpp->solve_system (top,gm,domproc,lhs,rhs,out);
    break;
  }
  default:{
    fprintf (stderr,"\n\n (%d) unknown domain decomposition solver is required",myrank);
    fprintf (stderr,"\n in function par_linear_solver (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }

}

/**
   function gathers contributions of boundary parts of local vectors
   (defined on subdomains) into one global vector (defined on boundaries)
   the global vector must be deleted outside of this subroutine
   
   @param lv - local vector
   @param gv - global vector, it is allocated only on the master processor
   
   JK, 6.11.2004
*/
void psolver::gather_bound_vect (double *lv,double *gv)
{
  switch (tdd){
  case schurcompldd:{
    schcom->gather_bound_vect (lv,gv,domproc);
    break;
  }
  default:{
    fprintf (stderr,"\n\n (%d) unknown domain decomposition solver is required",myrank);
    fprintf (stderr,"\n in function gather_bound_vect (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }

}

/**
   function scatters contributions from global vector (defined on boundaries)
   into local vectors (defined on subdomains)
   function adds contributions to the local vectors
   
   @param lv - local vector
   @param gv - global vector, it is allocated only on the master processor
   
   JK, 6.11.2004
*/
void psolver::scatter_bound_vect (double *lv,double *gv)
{
  switch (tdd){
  case schurcompldd:{
    schcom->scatter_bound_vect (lv,gv,domproc);
    break;
  }
  default:{
    fprintf (stderr,"\n\n (%d) unknown domain decomposition solver is required",myrank);
    fprintf (stderr,"\n in function gather_bound_vect (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }

}

/**
   function computes norm of vector of unbalanced values
   function gathers vector of unbalanced values on boundaries
   function scatters correct values into subdomains
   function returns norm of vector of unbalanced values
   
   @param lv - local vector
   
   JK, 6.11.2004
*/
double psolver::unbalanced_values (double *lv,FILE *out)
{
  double norm;
  
  switch (tdd){
  case schurcompldd:{
    norm = schcom->unbalanced_values (lv,domproc,out);
    break;
  }
  default:{
    fprintf (stderr,"\n\n (%d) unknown domain decomposition solver is required",myrank);
    fprintf (stderr,"\n in function gather_bound_vect (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  return norm;
}

/**
   function computes scalar product (dot product) in parallel
   
   @param lv - array containing %vector of appropriate subdomain
   @param out - output file (for auxiliary output)
   
   JK, 24.7.2008
*/
double psolver::pss (double *lv,FILE *out)
{
  long i,n;
  double dp=0.0;
  double *gv;
  MPI_Status stat;
  
  switch (tdd){
  case schurcompldd:{
    if (myrank==0){
      //  number of components of the global vector
      n=schcom->ndofcp;
      //  allocation of the global vector
      gv = new double [n];
      nullv (gv,n);
      //  assembling of the global vector
      //  dp contains contributions from the internal DOFs
      dp=schcom->pss_gather_bound_vect (lv,gv,domproc);
      //  computation of contributions from the interface DOFs
      for (i=0;i<n;i++){
	dp+=gv[i]*gv[i];
      }
      delete [] gv;
    }
    else{
      //  assembling of the global vector
      //  dp is useless on slaves, it is equal to zero
      dp=schcom->pss_gather_bound_vect (lv,gv,domproc);
    }
    //  scattering of scalar product of vector to the slaves
    if (myrank==0){
      for (i=1;i<nproc;i++){
        MPI_Send(&dp,1,MPI_DOUBLE,i,myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Recv (&dp,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    break;
  }
  default:{
    par_print_err(myrank,"unknown type of domain decomposition is required",__FILE__,__LINE__,__func__);
  }
  }
  
  return dp;
}




















/**
   function collects informations about computation
   it collects number of nodes, elements and stored %matrix entries on subdomains
   it searches for minimum and maximum of the previously mentioned numbers
   
   JK, 16.8.2007
*/
void psolver::computation_stat (gtopology *top,gmatrix *gm)
{
  long i,j,buffsize;
  long *buff;
  MPI_Status stat;
  
  buffsize = 3;
  buff = new long [3];

  //  number of nodes on subdomain
  buff[0]=top->nn;
  //  number of finite elements on subdomain
  buff[1]=top->ne;
  //  number of matrix entries
  buff[2]=gm->give_negm ();
  
  if (myrank==0){
    nndom = new long [nproc];
    nedom = new long [nproc];
    negmdom = new long [nproc];
    
    j=domproc[0];
    nndom[j]=buff[0];
    nedom[j]=buff[1];
    negmdom[j]=buff[2];
    
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      j=domproc[stat.MPI_TAG];
      nndom[j]=buff[0];
      nedom[j]=buff[1];
      negmdom[j]=buff[2];
    }
    
    minnn=LONG_MAX;
    maxnn=0;
    for (i=0;i<nproc;i++){
      if (minnn>nndom[i])
	minnn=nndom[i];
      if (maxnn<nndom[i])
	maxnn=nndom[i];
    }
    
    minne=LONG_MAX;
    maxne=0;
    for (i=0;i<nproc;i++){
      if (minne>nedom[i])
	minne=nedom[i];
      if (maxne<nedom[i])
	maxne=nedom[i];
    }
    
    minnegm=LONG_MAX;
    maxnegm=0;
    for (i=0;i<nproc;i++){
      if (minnegm>negmdom[i])
	minnegm=negmdom[i];
      if (maxnegm<negmdom[i])
	maxnegm=negmdom[i];
    }
    
  }
  else{
    MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  delete [] buff;
}


void psolver::computation_stat_print (FILE *out)
{
  long i;
  
  if (myrank==0){
    fprintf (out,"\n\n\n informations about computation \n");
    
    fprintf (out,"\n i   NN   NE    NEGM");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n %3ld %9ld  %9ld  %12ld",i+1,nndom[i],nedom[i],negmdom[i]);
    }

    fprintf (out,"\n");

    fprintf (out,"\n minimum number of nodes           %ld",minnn);
    fprintf (out,"\n maximum number of nodes           %ld",maxnn);

    fprintf (out,"\n minimum number of elements        %ld",minne);
    fprintf (out,"\n maximum number of elements        %ld",maxne);

    fprintf (out,"\n minimum number of matrix entries  %ld",minnegm);
    fprintf (out,"\n maximum number of matrix entries  %ld",maxnegm);
  }
}



/**
   Function compares two double values on master computer and scatters result to 
   slaves. It is used in nonlinear solvers for constraining common decision whether the
   error is in tolerance or not.
   
   @param val - compared value
   @param lim - limit to which the val is compared
   
   @retval  1 - in case that val > lim
   @retval  0 - in case that val == lim
   @retval -1 - in case that val < lim
   TKo, 31.7.2008
*/
long psolver::compare_on_master (double val, double lim)
{
  long ret = -2;
  long i;
  MPI_Status stat;

  if (myrank==0)
  {
    if (val > lim)
      ret = 1;
    else
    {
      if (val < lim)
        ret = -1;
      else
        ret = 0;
    }
    for (i=1;i<nproc;i++)
      MPI_Send(&ret,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
  }
  else
    MPI_Recv(&ret,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);

  MPI_Barrier (MPI_COMM_WORLD);
  
  return ret;
}
