#include "selnodes.h"
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "../GEFEL/galias.h"

/**
   constructor
   
   @param np - number of processors
   @param mr - my rank
   @param nd - number of subdomain
   @param k - number of nodes on subdomain
   @param j - array containing list of selected nodes
              it contains k components
   @param d - type of mesh description
   
   JK, 30.7.2007
*/
selnodes::selnodes(int np,int mr,long nd,long kk,long *j,meshdescription d,FILE *out,long mespr)
{
  long i,k;
  
  //  number of processors
  nproc=np;
  //  my rank
  myrank=mr;
  //  number of subdomain
  ndom=nd;

  //  mesh description
  md = d;
  
  //  number of nodes
  nn=kk;


  //  number of selected nodes
  nsn=0;
  for (i=0;i<nn;i++){
    if (j[i]>-1)
      nsn++;
  }
  
  if (nsn==0){
    print_err("wrong number of selected nodes in constructor", __FILE__, __LINE__, __func__);
    abort();
  }
  
  //  list of selected nodes - local numbers
  lsnl=new long [nsn];
  //  list of selected nodes - global numbers
  lsng=new long [nsn];
  
  k=0;
  for (i=0;i<nn;i++){
    if (j[i]>-1){
      lsnl[k]=i;
      lsng[k]=j[i];
      k++;
    }
  }
  
  if (mespr==0){
    fprintf (out,"\n\n\n list of local and global/group numbers of selected nodes (lsnl, lsng)\n");
    fprintf (out,"\n number of selected nodes is %ld",nsn);
    for (i=0;i<nsn;i++){
      fprintf (out,"\n local node number %6ld       global/group node number %6ld",lsnl[i],lsng[i]);
    }
  }

  //  number of all DOFs including prescribed DOFs
  ndof=0;
  //  maximum number of selected nodes on subdomain
  maxnsn=0;
  //  maximum number of DOFs on subdomain
  maxndof=0;
  //  maximum number of contributing nodes on subdomain in the FETI method
  maxncn=0;
  
  ncn=0;

  ldof = NULL;
  
  ldofmultip = NULL;
  if (myrank==0){
    //  total number of selected nodes
    tnsn=0;
    
    //  numbers of selected nodes
    nsndom = NULL;
    //  group node numbers
    gnn = NULL;
    //  node multiplicty
    nodmultip = NULL;
    // dof multiplicity
    dofmultip = NULL; 
    //  numbers of DOFs
    ndofdom = NULL;
    //  number of DOFs on nodes
    ndofnmas = NULL;
    //  DOFs or indicators on master
    dofmas = NULL;
    //  numbers of DOFs for selected nodes
    ndofnsn = NULL;
    //  code numbers at selected nodes on master
    cnm = NULL;
    //  code numbers on master
    cndom = NULL;
    
    ncndom = NULL;
  }
  
}

/**
   destructor
   
   JK, 30.7.2007
*/
selnodes::~selnodes()
{
  long i;

  delete [] lsnl;
  delete [] lsng;
  
  if (myrank==0){
    delete [] nsndom;
    delete [] ndofdom;
    
    for (i=0;i<nproc;i++){
      delete [] gnn[i];
    }
    delete [] gnn;

    for (i=0;i<nproc;i++){
      delete [] ndofnmas[i];
    }
    delete [] ndofnmas;
  }
}

/**
   function computes number of selected nodes and assembles
   array on master
   
   function assembles array nsndom

   maxnsn is obtained

   30.7.2007, JK
*/
void selnodes::number_of_selected_nodes (long *domproc,FILE *out)
{
  long i,j,k;
  MPI_Status stat;
  
  if (myrank==0){
    //  array containing numbers of selected nodes on subdomains
    if (nsndom!=NULL)
      delete [] nsndom;
    nsndom = new long [nproc];
    
    //  maximum number of selected nodes on subdomain
    maxnsn=nsn;

    //  master contributions
    j=domproc[0];
    nsndom[j]=nsn;
    
    for (i=1;i<nproc;i++){
      MPI_Recv (&k,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      //  slave contributions
      j=domproc[stat.MPI_TAG];
      nsndom[j]=k;
      if (maxnsn<k)
	maxnsn=k;
    }
    
    for (i=1;i<nproc;i++){
      MPI_Send (&maxnsn,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
  }
  else{
    MPI_Send (&nsn,1,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    MPI_Recv (&maxnsn,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  

  if (myrank==0){
    fprintf (out,"\n\n numbers of selected nodes on subdomains (nsndom)\n");
    fprintf (out,"\n maximum number of selected nodes on subdomain  %ld\n",maxnsn);
    for (i=0;i<nproc;i++){
      fprintf (out,"\n subdomain %6ld     %ld",i+1,nsndom[i]);
    }
  }

}

/**
   function collects node numbers on master
   
   function assembles array gnn
   
   @param tnnp - total number of nodes in problem
   
   30.7.2007, JK
*/
void selnodes::nodes_on_master (long *domproc,FILE *out)
{
  long i,j,k,max,tnnp,buffsize;
  long *buff,*nl;
  MPI_Status stat;
  
  buffsize=maxnsn;
  buff = new long [buffsize];
  for (i=0;i<nsn;i++){
    buff[i]=lsng[i];
  }
  
  switch (md){
  case all_nodes:
  case neg_bound_nodes:{
    
    max=0;
    for (i=0;i<nsn;i++){
      if (max<lsng[i])
	max=lsng[i];
    }
    
    if (myrank==0){
      for (i=1;i<nproc;i++){
	MPI_Recv (&k,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	if (max<k)
	  max=k;
      }
      tnnp=max;
      tnnp++;
    }
    else{
      MPI_Send (&max,1,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    }
    MPI_Barrier (MPI_COMM_WORLD);


    
    if (myrank==0){
      //  group node numbers on master
      gnn = new long* [nproc];
      for (i=0;i<nproc;i++){
	gnn[i] = new long [nsndom[i]];
      }
      
      nl=new long [tnnp];
      for (i=0;i<tnnp;i++){
	nl[i]=-1;
      }
      
      //  master contribution
      k=domproc[0];
      for (j=0;j<nsndom[k];j++){
	nl[buff[j]]++;
	gnn[k][j]=buff[j];
      }
      
      //  slave contribution
      for (i=1;i<nproc;i++){
	MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	k=domproc[stat.MPI_TAG];
	for (j=0;j<nsndom[k];j++){
	  nl[buff[j]]++;
	  gnn[k][j]=buff[j];
	}
      }
      
      //  total number of selected nodes
      //  generation of new node numbers
      tnsn=0;
      for (i=0;i<tnnp;i++){
	if (nl[i]>-1){
	  nl[i]=tnsn;
	  tnsn++;
	}
      }

      for (i=0;i<nproc;i++){
	for (j=0;j<nsndom[i];j++){
	  gnn[i][j]=nl[gnn[i][j]];
	}
      }
      
      delete [] nl;
    }
    else{
      MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    break;
  }
    
  case bound_nodes:{
    
    if (myrank==0){
      //  group node numbers on master
      gnn = new long* [nproc];
      for (i=0;i<nproc;i++){
	gnn[i] = new long [nsndom[i]];
      }
      
      //  total number of selected nodes
      tnsn=0;
      
      //  master contribution
      k=domproc[0];
      for (j=0;j<nsndom[k];j++){
	gnn[k][j]=buff[j];
	if (tnsn<buff[j])
	  tnsn=buff[j];
      }

      //  slave contribution
      for (i=1;i<nproc;i++){
	MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	k=domproc[stat.MPI_TAG];
	for (j=0;j<nsndom[k];j++){
	  gnn[k][j]=buff[j];
	  if (tnsn<buff[j])
	    tnsn=buff[j];
	}
      }
      tnsn++;
    }
    else{
      MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    break;
  }
  default:{
    fprintf (stderr,"\n\n wrong type of mesh description is required in functio");
    fprintf (stderr,"\n node_on_master (file %s, line %d)\n",__FILE__,__LINE__);
  }
  }
  

  if (myrank==0){
    fprintf (out,"\n\n total number of selected nodes is %ld",tnsn);
    fprintf (out,"\n\n group node numbers (gnn)\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n\n subdomain %6ld",i+1);
      for (j=0;j<nsndom[i];j++){
	fprintf (out,"\n %6ld   %ld",j+1,gnn[i][j]);
      }
    }
  }

  delete [] buff;
}

/**
   function computes multiplicity of selected nodes
   
   function assembles array nodmultip
   
   JK, 31.7.2007
*/
void selnodes::node_multiplicity (FILE *out)
{
  long i,j;
  
  if (myrank==0){
    if (nodmultip!=NULL)
      delete [] nodmultip;
    nodmultip = new long [tnsn];
    for (i=0;i<tnsn;i++){
      nodmultip[i]=0;
    }
    
    for (i=0;i<nproc;i++){
      for (j=0;j<nsndom[i];j++){
	nodmultip[gnn[i][j]]++;
      }
    }
    

    fprintf (out,"\n\n node multiplicity (nodmultip)\n");
    for (i=0;i<tnsn;i++){
      fprintf (out,"\n selected node %6ld    %ld",i+1,nodmultip[i]);
    }

  }
}

void selnodes::dof_multiplicity (FILE *out)
{
  long i,j,k,n,nm,ndofn;
  MPI_Status stat;
  long buffsize;
  long *buff;
  
  buffsize = maxndof+1;
  buff = new long [buffsize];

  if (myrank == 0){
    if(dofmultip != NULL){
      delete []dofmultip;
    }
    
    
    dofmultip = new long[tndof];
    n = 0;
    for(i = 0; i < tnsn; i++){
      nm = nodmultip[i];
      for(j = 0; j < nm-1; j++){
	ndofn = ndofnsn[i];
	for(k = 0; k < ndofn; k++){
	  if(doffeti[i][j][k] > 0){
	    dofmultip[n] = nm;
	    n++;
	  }
	}
      }
    }
    // pro ladeni
    fprintf (out,"\n\n\n Control of array dofmultip\n\n\nMultiplicity of coarse DOF n = %ld\n\n",n); 
    for(i = 0; i < tndof; i++){
      fprintf (out,"coarse DOF %ld has multiplicity %ld\n",i+1,dofmultip[i]); 
    }
    /*
    for(i = 1; i < nproc; i++){
      
      for(j = 0; j < ndofdom[i];j++){
	//fprintf (out,"cndom = %ld \n",cndom[i][j]);
	buff[j] = dofmultip[cndom[i][j]-1];
      }
      buff[maxndof] = ndofdom[i];
      fprintf (out,"ndofdom[%ld] = %ld %ld\n",i,ndofdom[i],buff[maxndof]);
      MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD); 
    }
    
    ldofmultip = new long [ndofdom[0]];
    for(j = 0; j < ndofdom[0];j++){
      ldofmultip[j] = dofmultip[cndom[0][j]-1];
    }
    
    // pro ladeni
    fprintf (out,"ndof = %ld\n",ndofdom[0]);
    for(i = 0; i < ndofdom[0]; i++){
      fprintf (out,"ldofmultip %ld\n",ldofmultip[i]); 
    }
    */
  }
  /*
  else{
    MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    if(ldofmultip != NULL){
      delete []dofmultip;
    }
    ldofmultip = new long [ndof];
    for(i = 0; i < ndof; i++){
      ldofmultip[i] = buff[i];
    }
       // pro ladeni
    fprintf (out,"ndof = %ld\n",ndof);
    for(i = 0; i < ndof; i++){
      fprintf (out," ldofmultipl %ld\n",ldofmultip[i]); 
    }
   
  }
  MPI_Barrier (MPI_COMM_WORLD);
  delete []buff;
  */
}


/**
   function searches for all possible DOFs in selected nodes
   prescribed DOFs are included
   
   maxndof is determined
   array ndofdom is assembled
   
   JK, 31.7.2007
*/
void selnodes::number_all_dofs (gtopology *top,long *domproc,FILE *out)
{
  long i,j,k;
  MPI_Status stat;
  
  ndof=0;
  for (i=0;i<nsn;i++){
    j=lsnl[i];
    ndof+=top->give_ndofn (j);
  }
  
  if (myrank==0){
    //  number of all dofs on subdomains
    if (ndofdom!=NULL)
      delete [] ndofdom;
    ndofdom=new long [nproc];
    
    //  maximum number of all dofs on subdomains
    maxndof=ndof;
    
    //  master contributions
    j=domproc[0];
    ndofdom[j]=ndof;

    for (i=1;i<nproc;i++){
      MPI_Recv (&k,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      //  slave contributions
      j=domproc[stat.MPI_TAG];
      ndofdom[j]=k;
      if (maxndof<k)
	maxndof=k;
    }

    for (i=1;i<nproc;i++){
      MPI_Send (&maxndof,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
    
    
  }
  else{
    MPI_Send (&ndof,1,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    MPI_Recv (&maxndof,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  

  if (myrank==0){
    fprintf (out,"\n\n numbers of DOFs on subdomains (ndofdom)\n");
    fprintf (out,"\n maximum number of DOFs on subdomains is %ld\n",maxndof);
    for (i=0;i<nproc;i++){
      fprintf (out,"\n subdomain number %6ld   %ld",i+1,ndofdom[i]);
    }
  }

}

/**
   function collects numbers of DOFs on selected nodes
   numbers are stored on master in the array ndofnmas
   
   array ndofnmas is assembled

   JK, 31.7.2007
*/
void selnodes::ndofn_on_master (gtopology *top,long *domproc,FILE *out)
{
  long i,j,k,buffsize;
  long *buff;
  MPI_Status stat;

  buffsize=maxndof;
  buff = new long [buffsize];
  for (i=0;i<nsn;i++){
    j=lsnl[i];
    buff[i]=top->give_ndofn (j);
  }
  
  if (myrank==0){
    //  group node numbers on master
    ndofnmas = new long* [nproc];
    for (i=0;i<nproc;i++){
      ndofnmas[i] = new long [nsndom[i]];
    }
    
    //  master contribution
    k=domproc[0];
    for (j=0;j<nsndom[k];j++){
      ndofnmas[k][j]=buff[j];
    }
    
    //  slave contribution
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      k=domproc[stat.MPI_TAG];
      for (j=0;j<nsndom[k];j++){
	ndofnmas[k][j]=buff[j];
      }
    }
    
  }
  else{
    MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  delete [] buff;
  

  if (myrank==0){
    fprintf (out,"\n\n numbers of DOFs on selected nodes (ndofnmas)\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n subdomain number %6ld",i+1);
      for (j=0;j<nsndom[i];j++){
	fprintf (out,"\n selected node %6ld   %ld",j+1,ndofnmas[i][j]);
      }
    }
  }

}


/**
   function assembles DOFs indicators on master
   
   array dofmas is assembled
   
   @param top - pointer to the genral topology
   @param domproc - correspondence between processors and subdomains
   @param out - output file (for auxiliary output)
   
   JK, 31.7.2007
*/
void selnodes::dof_indicators (gtopology *top,long *domproc,FILE *out)
{
  long i,j,k,l,m,ndofn,buffsize;
  long *buff;
  MPI_Status stat;

  buffsize=maxndof;
  buff = new long [buffsize];
  l=0;
  for (i=0;i<nsn;i++){
    j=lsnl[i];
    ndofn=top->give_ndofn (j);
    for (k=0;k<ndofn;k++){
      buff[l]=top->give_dof (j,k);
      l++;
    }
  }
  
  
  if (myrank==0){
    
    dofmas = new long** [nproc];
    for (i=0;i<nproc;i++){
      dofmas[i] = new long* [nsndom[i]];
      for (j=0;j<nsndom[i];j++){
	dofmas[i][j] = new long [ndofnmas[i][j]];
      }
    }
    
    //  master contribution
    k=domproc[0];
    m=0;
    for (j=0;j<nsndom[k];j++){
      for (l=0;l<ndofnmas[k][j];l++){
	dofmas[k][j][l]=buff[m];
	m++;
      }
    }
    
    //  slave contribution
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      k=domproc[stat.MPI_TAG];
      m=0;
      for (j=0;j<nsndom[k];j++){
	for (l=0;l<ndofnmas[k][j];l++){
	  dofmas[k][j][l]=buff[m];
	  m++;
	}
      }
    }

  }
  else{
    MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);

  delete [] buff;
  

  if (myrank==0){
    fprintf (out,"\n\n DOFs indicators on master (dofmas)\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n subdomain number %6ld",i+1);
      for (j=0;j<nsndom[i];j++){
	fprintf (out,"\n selected node %6ld    ",j+1);
	for (k=0;k<ndofnmas[i][j];k++){
	  fprintf (out,"  %ld",dofmas[i][j][k]);
	}
      }
    }
  }

}

/**
   function generates ordering of selected nodes with respect
   to the Schur complement method
   
   the function generates code numbers / numbers of unknowns
   on the coarse problem
   
   array ldof is assembled

   @param top - pointer to the general topology
   @param out - output file (for auxiliary output)
   
   JK, 31.7.2007
*/
void selnodes::schur_ordering (gtopology *top,FILE *out)
{
  long i,j,k,l,m,ndofn;
  
  if (myrank==0){
    //  determination of numbers of DOFs at selected nodes
    if (ndofnsn!=NULL)
      delete [] ndofnsn;
    ndofnsn = new long [tnsn];
    for (i=0;i<tnsn;i++){
      ndofnsn[i]=0;
    }
    
    for (i=0;i<nproc;i++){
      for (j=0;j<nsndom[i];j++){
	k=ndofnmas[i][j];
	l=gnn[i][j];
	if (ndofnsn[l]>0){
	  if (ndofnsn[l]!=k){
	    fprintf (stderr,"\n\n incompatible number of DOFs at selected node number %ld (file %s, line %d)\n",l,__FILE__,__LINE__);
	  }
	}
	else{
	  ndofnsn[l]=k;
	}
      }
    }
    
    //fprintf (out,"\n\n kontrola ndofnsn");
    //for (i=0;i<tnsn;i++){
    //fprintf (out,"\n %ld   %ld",i+1,ndofnsn[i]);
    //}


    //  code number indicators
    cnm = new long* [tnsn];
    for (i=0;i<tnsn;i++){
      cnm[i] = new long [ndofnsn[i]];
    }
    
    for (i=0;i<nproc;i++){
      for (j=0;j<nsndom[i];j++){
	l=gnn[i][j];
	for (k=0;k<ndofnsn[l];k++){
	  cnm[l][k]=dofmas[i][j][k];
	}
      }
    }
    

    //  code number generation
    tndof=1;
    for (i=0;i<tnsn;i++){
      for (j=0;j<ndofnsn[i];j++){
	if (cnm[i][j]>0){
	  cnm[i][j]=tndof;
	  tndof++;
	}
      }
    }
    tndof--;
    
    
    //  computation of real number of DOFs on subdomains
    //  supports are not included now
    for (i=0;i<nproc;i++){
      ndofdom[i]=0;
      for (j=0;j<nsndom[i];j++){
	l=gnn[i][j];
	for (k=0;k<ndofnsn[l];k++){
	  if (cnm[l][k]>0)
	    ndofdom[i]++;
	}
      }
    }
    
    cndom = new long *[nproc];
    for (i=0;i<nproc;i++){
      cndom[i] = new long [ndofdom[i]];
    }
    
    for (i=0;i<nproc;i++){
      m=0;
      for (j=0;j<nsndom[i];j++){
	l=gnn[i][j];
	for (k=0;k<ndofnsn[l];k++){
	  if (cnm[l][k]>0){
	    cndom[i][m]=cnm[l][k];
	    m++;
	  }
	}
      }
    }


  }
  
  //  zde by se dalo smazat pole cnm



  

  //  number of contributions to the coarse problem
  //  number of DOFs which contribute to the coarse problem
  ndof=0;
  for (i=0;i<nsn;i++){
    l=lsnl[i];
    ndofn=top->give_ndofn (l);
    for (k=0;k<ndofn;k++){
      if (top->give_dof (l,k)>0)
	ndof++;
    }
  }
  
  //  list of code numbers which are extracted from domain in the FETI method
  //  some code numbers are positive and some of them are negative
  //  it depends on number of subdomain
  //  these code numbers are used on subdomains, master processor contains corresponding array with coarse code numbers
  
  if (ldof!=NULL)
    delete [] ldof;
  ldof = new long [ndof];
  
  m=0;
  for (i=0;i<nsn;i++){
    l=lsnl[i];
    ndofn=top->give_ndofn (l);
    for (j=0;j<ndofn;j++){
      k=top->give_dof (l,j);
      if (k>0){
	ldof[m]=k;
	m++;
      }
    }
  }













  /*
  fprintf (out,"\n array ldof");
  for (i=0;i<ndof;i++){
    fprintf (out,"\n %6ld   %ld",i+1,ldof[i]);
  }
  
  
  if (myrank==0){
    fprintf (out,"\n\n actual numbers of DOFs on subdomains (ndofdom)\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n subdomain number %6ld   %ld",i+1,ndofdom[i]);
    }
    
    fprintf (out,"\n\n code numbers of Schur ordering (cndom)\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n subdomain number %6ld",i+1);
      for (j=0;j<ndofdom[i];j++){
	fprintf (out,"\n dof %6ld   %ld",j+1,cndom[i][j]);
      }
    }
  }
  */
}




/**
   function assembles lists of connected nodes to coarse nodes
   and numbers of subdomains which contain connected nodes
   
   arrays ljn and lsn are assembled
   
   JK, 1.8.2007
*/
void selnodes::group_local_nodes (FILE *out)
{
  long i,j,k,m,n,min;
  
  if (myrank==0){
    ljn = new long* [tnsn];
    lsn = new long* [tnsn];
    for (i=0;i<tnsn;i++){
      ljn[i] = new long [nodmultip[i]];
      lsn[i] = new long [nodmultip[i]];
      nodmultip[i]=0;
    }
    
    //  list of local node numbers of connected nodes
    //  list of numbers of subdomains which contains connected nodes
    for (i=0;i<nproc;i++){
      for (j=0;j<nsndom[i];j++){
	k=gnn[i][j];
	ljn[k][nodmultip[k]]=j;
	lsn[k][nodmultip[k]]=i;
	nodmultip[k]++;
      }
    }
    
    
    //  node sorting
    for (i=0;i<tnsn;i++){
      for (j=0;j<nodmultip[i];j++){
	min=nproc;
	for (k=j;k<nodmultip[i];k++){
	  if (lsn[i][k]<min){
	    min=lsn[i][k];
	    m=k;
	  }
	}
	n=lsn[i][j];
	lsn[i][j]=lsn[i][m];
	lsn[i][m]=n;

	n=ljn[i][j];
	ljn[i][j]=ljn[i][m];
	ljn[i][m]=n;
      }
    }

  }
  

  if (myrank==0){
    fprintf (out,"\n\n list of local node numbers of connected nodes (ljn)\n");
    fprintf (out,"\n\n list of numbers of subdomains which contains connected nodes (lsn)\n");
    for (i=0;i<tnsn;i++){
      fprintf (out,"\n coarse node %6ld   multip %6ld   ",i+1,nodmultip[i]);
      for (j=0;j<nodmultip[i];j++){
	fprintf (out,"    %ld %ld",ljn[i][j],lsn[i][j]);
      }
    }
  }

}

/**
   function assembles indicators of code numbers and then generates
   code numbers
   
   array doffeti is assembled

   JK, 1.8.2007
*/
void selnodes::dof_feti (FILE *out)
{
  long i,j,k,l,m;
  
  if (myrank==0){

    //  determination of numbers of DOFs at selected nodes
    if (ndofnsn!=NULL)
      delete [] ndofnsn;
    ndofnsn = new long [tnsn];
    for (i=0;i<tnsn;i++){
      ndofnsn[i]=0;
    }
    
    for (i=0;i<nproc;i++){
      for (j=0;j<nsndom[i];j++){
	k=ndofnmas[i][j];
	l=gnn[i][j];
	if (ndofnsn[l]>0){
	  if (ndofnsn[l]!=k){
	    fprintf (stderr,"\n\n incompatible number of DOFs at selected node number %ld (file %s, line %d)\n",l,__FILE__,__LINE__);
	  }
	}
	else{
	  ndofnsn[l]=k;
	}
      }
    }
    
    //fprintf (out,"\n\n kontrola ndofnsn");
    //for (i=0;i<tnsn;i++){
    //fprintf (out,"\n %ld   %ld",i+1,ndofnsn[i]);
    //}
    
    
    //  array allocation
    doffeti = new long** [tnsn];
    for (i=0;i<tnsn;i++){
      doffeti[i] = new long* [nodmultip[i]];
      for (j=0;j<nodmultip[i];j++){
	doffeti[i][j] = new long [ndofnsn[i]];
      }
    }
    
    //  assembling of code numbers indicators
    for (i=0;i<tnsn;i++){
      for (j=0;j<nodmultip[i];j++){
	l=ljn[i][j];
	m=lsn[i][j];
	for (k=0;k<ndofnsn[i];k++){
	  doffeti[i][j][k]=dofmas[m][l][k];
	}
      }
    }
    
    //  code numbers generation
    tndof=1;
    for (i=0;i<tnsn;i++){
      for (j=0;j<nodmultip[i]-1;j++){
	for (k=0;k<ndofnsn[i];k++){
	  if (doffeti[i][j][k]>0){
	    doffeti[i][j][k]=tndof;
	    tndof++;
	  }
	}
      }
    }
    tndof--;
  }
  

  if (myrank==0){
    fprintf (out,"\n\n code numbers for FETI method (doffeti)\n");
    for (i=0;i<tnsn;i++){
      fprintf (out,"\n coarse node %6ld",i+1);
      for (j=0;j<nodmultip[i]-1;j++){
	fprintf (out,"  con.n. %6ld  sub.n. %6ld   ",ljn[i][j],lsn[i][j]);
	for (k=0;k<ndofnsn[i];k++){
	  fprintf (out," %ld",doffeti[i][j][k]);
	}
      }
    }
  }

}

/**
   function determines number of contributions to coarse problem
   from particular subdomains
   
   maxndof is rewritten
   ndofdom is rewritten
   
   array ncn is assembled
   
   JK, 1.8.2007
*/
void selnodes::number_contrib (long *domproc,FILE *out)
{
  long i,j,k,l,m,ii,buffsize;
  long *buff;
  MPI_Status stat;
  
  buffsize=3;
  buff = new long [buffsize];
  
  if (myrank==0){

    //  allocation of array
    if (ndofdom!=NULL)
      delete [] ndofdom;
    ndofdom = new long [nproc];
    for (i=0;i<nproc;i++){
      ndofdom[i]=0;
    }
    
    if (ncndom!=NULL)
      delete [] ncndom;
    ncndom = new long [nproc];
    for (i=0;i<nproc;i++){
      ncndom[i]=0;
    }
     
    //  number of contributing DOFs from subdomains to coarse problem
    for (i=0;i<nproc;i++){
      for (j=0;j<nsndom[i];j++){
	ii=gnn[i][j];
	for (k=0;k<nodmultip[ii];k++){
	  if (lsn[ii][k]==i){
	    if (k==0){
	      for (l=0;l<nodmultip[ii]-1;l++){
		ncndom[i]++;
		for (m=0;m<ndofnsn[ii];m++){
		  if (doffeti[ii][l][m]>0)
		    ndofdom[i]++;
		}
	      }
	    }
	    else{
	      ncndom[i]++;
	      for (m=0;m<ndofnsn[ii];m++){
		if (doffeti[ii][k-1][m]>0)
		  ndofdom[i]++;
	      }
	    }
	    break;
	  }
	}
      }
    }
    
    //  maximum number of contributions from subdomain
    maxndof=0;
    //  maximum number of contributing nodes from one subdomain
    maxncn=0;
    for (i=0;i<nproc;i++){
      if (maxndof<ndofdom[i])
	maxndof=ndofdom[i];
      if (maxncn<ncndom[i])
	maxncn=ncndom[i];
    }
    
    buff[0]=maxndof;
    buff[1]=maxncn;
    
    for (i=1;i<nproc;i++){
      j=domproc[i];
      buff[2]=ncndom[j];
      MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
    
    j=domproc[0];
    buff[2]=ncndom[j];
  }
  else{
    MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  maxndof=buff[0];
  maxncn=buff[1];
  ncn=buff[2];
  

  if (myrank==0){
    fprintf (out,"\n\n number of contributing DOFs to coarse problem (ndofdom)\n");
    fprintf (out,"\n maximum number of DOFs on subdomain %ld",maxndof);
    fprintf (out,"\n maximum number of contributing nodes on subdomain %ld",maxncn);
    for (i=0;i<nproc;i++){
      fprintf (out,"\n subdomain number %6ld   %ld",i+1,ndofdom[i]);
    }
  }
  fprintf (out,"\n\n number of nodes which contribute to coarse FETI problem (ncn) %ld",ncn);

}


/**
   function assmebles nodes contributing to the coarse proble in the FETI method
      
   JK, 1.8.2007
*/
void selnodes::contrib_dofs (gtopology *top,long *domproc,FILE *out)
{
  long i,j,k,l,m,ii,jj,ndofn,buffsize,min;
  long *buff;
  MPI_Status stat;
  
  buffsize=maxncn;
  buff = new long [buffsize];
  
  if (myrank==0){
    
    //  array of local numbers of nodes contributing to the coarse problem
    for (i=1;i<nproc;i++){
      
      for (j=0;j<buffsize;j++){
	buff[j]=0;
      }
      
      m=0;
      for (j=0;j<tnsn;j++){
	for (k=0;k<nodmultip[j];k++){
	  if (lsn[j][k]==domproc[i]){
	    if (k==0){
	      for (l=0;l<nodmultip[j]-1;l++){
		buff[m]=ljn[j][k];
		buff[m]=0-buff[m]-1;
		m++;
	      }
	    }
	    else{
	      buff[m]=ljn[j][k];
	      m++;
	    }
	    break;
	  }
	}
      }
      
      MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
    
    i=0;
    for (j=0;j<buffsize;j++){
      buff[j]=0;
    }
    
    m=0;
    for (j=0;j<tnsn;j++){
      for (k=0;k<nodmultip[j];k++){
	if (lsn[j][k]==domproc[i]){
	  if (k==0){
	    for (l=0;l<nodmultip[j]-1;l++){
	      buff[m]=ljn[j][k];
	      buff[m]=0-buff[m]-1;
	      m++;
	    }
	  }
	  else{
	    buff[m]=ljn[j][k];
	    m++;
	  }
	  break;
	}
      }
    }
    
  }
  else{
    MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  

  fprintf (out,"\n\n kontrola bufferu na vsech uzlech");
  for (i=0;i<ncn;i++){
    fprintf (out,"\n %ld   %ld",i+1,buff[i]);
  }


  for (i=0;i<ncn;i++){
    if (buff[i]<0)
      min=0-buff[i]-1;
    else
      min=buff[i];
      
    l=i;
    for (j=i+1;j<ncn;j++){
      if (buff[j]<0)
	k=0-buff[j]-1;
      else
	k=buff[j];
      
      if (min>k){
	l=j;
	min=k;
      }
    }
    
    
    k=buff[i];
    buff[i]=buff[l];
    buff[l]=k;
  }

  fprintf (out,"\n\n kontrola bufferu na vsech uzlech");
  for (i=0;i<ncn;i++){
    fprintf (out,"\n %ld   %ld",i+1,buff[i]);
  }
  fprintf (out,"\n konec kontroly\n");
  /*
  fprintf (out,"\n\n kontrola lsnl na vsech uzlech");
  for (i=0;i<nsn;i++){
    fprintf (out,"\n %ld   %ld",i+1,lsnl[i]);
  }
  fprintf (out,"\n konec kontroly\n");
  */
  
  
  //  number of contributions to the coarse problem
  //  number of DOFs which contribute to the coarse problem
  ndof=0;
  for (i=0;i<ncn;i++){
    j=buff[i];
    if (j<0){
      j=0-j-1;
    }
    l=lsnl[j];
    ndofn=top->give_ndofn (l);
    for (k=0;k<ndofn;k++){
      if (top->give_dof (l,k)>0)
	ndof++;
    }
  }
  
  //  list of code numbers which are extracted from domain in the FETI method
  //  some code numbers are positive and some of them are negative
  //  it depends on number of subdomain
  //  these code numbers are used on subdomains, master processor contains corresponding array with coarse code numbers

  if (ldof!=NULL)
    delete [] ldof;
  ldof = new long [ndof];
  
  m=0;
  for (i=0;i<ncn;i++){
    ii=buff[i];
    if (ii<0){
      ii=0-ii-1;
      jj=lsnl[ii];
      ndofn=top->give_ndofn (jj);
      for (j=0;j<ndofn;j++){
	l=top->give_dof (jj,j);
	if (l>0){
	  ldof[m]=0-l;
	  m++;
	}
      }
    }
    else{
      jj=lsnl[ii];
      ndofn=top->give_ndofn (jj);
      for (j=0;j<ndofn;j++){
	l=top->give_dof (jj,j);
	if (l>0){
	  ldof[m]=l;
	  m++;
	}
      }
    }
  }
  
  delete [] buff;
  
  
  if (myrank==0){
    cndom = new long* [nproc];
    for (i=0;i<nproc;i++){
      cndom[i] = new long [ndofdom[i]];
    }
    
    for (i=0;i<nproc;i++){
      l=0;
      for (j=0;j<nsndom[i];j++){
	m=gnn[i][j];
	for (k=0;k<nodmultip[m];k++){
	  if (lsn[m][k]==i){
	    if (k==0){
	      for (ii=0;ii<nodmultip[m]-1;ii++){
		for (jj=0;jj<ndofnsn[m];jj++){
		  if (doffeti[m][ii][jj]>0){
		    cndom[i][l]=doffeti[m][ii][jj];
		    l++;
		  }
		}
	      }
	    }
	    else{
	      for (jj=0;jj<ndofnsn[m];jj++){
		if (doffeti[m][k-1][jj]>0){
		  cndom[i][l]=doffeti[m][k-1][jj];
		  l++;
		}
	      }
	    }
	    break;
	  }
	}
      }
    }

  }
  
  

  fprintf (out,"\n\n number of DOFs which contribute to coarse FETI problem %ld",ndof);
  fprintf (out,"\n array ldof");
  for (i=0;i<ndof;i++){
    fprintf (out,"\n %6ld   %ld",i+1,ldof[i]);
  }
  
  if (myrank==0){
    fprintf (out,"\n\n code numbers on master (cndom)\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n subdomain number %6ld (number of DOFs %ld)",i+1,ndofdom[i]);
      for (j=0;j<ndofdom[i];j++){
	fprintf (out,"\n %6ld   %ld",j+1,cndom[i][j]);
      }
    }
  }

}

